@@ -1,2 +1,2 @@
 from memorizing_transformers_pytorch.memorizing_transformers_pytorch import MemorizingTransformer
-from memorizing_transformers_pytorch.ann_memory import ANNMemory
+from memorizing_transformers_pytorch.knn_memory import KNNMemory
