@@ -11,6 +11,10 @@ from contextlib import ExitStack, contextmanager
 from einops import rearrange
 from einops_exts import rearrange_with_anon_dims, check_shape
 
+# multiprocessing
+
+from joblib import Parallel, delayed, cpu_count
+
 # constants
 
 FAISS_INDEX_GPU_ID = int(os.getenv('FAISS_INDEX_GPU_ID', 0))
@@ -50,15 +54,14 @@ class KNN():
         max_num_entries,
         cap_num_entries = False,
         M = 15,
-        use_gpu = False
+        keep_stats = False
     ):
         index = faiss.IndexHNSWFlat(dim, M, faiss.METRIC_INNER_PRODUCT)
-        self.use_gpu = use_gpu
-
-        self.index = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(), FAISS_INDEX_GPU_ID, index) if use_gpu else index
+        self.index = index
         self.max_num_entries = max_num_entries
         self.cap_num_entries = cap_num_entries
         self.is_trained = False
+        self.keep_stats = keep_stats
 
         self.reset()
 
@@ -68,9 +71,12 @@ class KNN():
 
     def reset(self):
         self.ids = np.empty((0,), dtype = np.int32)
-        self.hits = np.empty((0,), dtype = np.int32)
-        self.age_num_iterations = np.empty((0,), dtype = np.int32)
-        self.ages_since_last_hit = np.empty((0,), dtype = np.int32)
+
+        if self.keep_stats:
+            self.hits = np.empty((0,), dtype = np.int32)
+            self.age_num_iterations = np.empty((0,), dtype = np.int32)
+            self.ages_since_last_hit = np.empty((0,), dtype = np.int32)
+
         self.index.reset()
         self.is_trained = False
 
@@ -83,9 +89,11 @@ class KNN():
             self.train(x)
 
         self.ids = np.concatenate((ids, self.ids))
-        self.hits = np.concatenate((np.zeros_like(ids), self.hits))
-        self.age_num_iterations = np.concatenate((np.zeros_like(ids), self.age_num_iterations))
-        self.ages_since_last_hit = np.concatenate((np.zeros_like(ids), self.ages_since_last_hit))
+
+        if self.keep_stats:
+            self.hits = np.concatenate((np.zeros_like(ids), self.hits))
+            self.age_num_iterations = np.concatenate((np.zeros_like(ids), self.age_num_iterations))
+            self.ages_since_last_hit = np.concatenate((np.zeros_like(ids), self.ages_since_last_hit))
 
         if self.cap_num_entries and len(self.ids) > self.max_num_entries:
             self.reset()
@@ -106,7 +114,7 @@ class KNN():
 
         distances, indices = self.index.search(x, k = topk)
 
-        if increment_hits:
+        if increment_hits and self.keep_stats:
             hits = count_intersect(self.ids, rearrange(indices, '... -> (...)'))
             self.hits += hits
 
@@ -131,7 +139,7 @@ class KNNMemory():
         max_memories = 16000,
         num_indices = 1,
         memmap_filename = './knn.memory.memmap',
-        knn_use_gpu = False
+        multiprocessing = True
     ):
         self.dim = dim
         self.num_indices = num_indices
@@ -142,8 +150,10 @@ class KNNMemory():
         self.db_offsets = np.zeros(num_indices, dtype = np.int32)
 
         self.db = np.memmap(memmap_filename, mode = 'w+', dtype = np.float32, shape = self.shape)
-        self.knns = [KNN(dim = dim, max_num_entries = max_memories, use_gpu = knn_use_gpu, cap_num_entries = True) for _ in range(num_indices)]
+        self.knns = [KNN(dim = dim, max_num_entries = max_memories, cap_num_entries = True) for _ in range(num_indices)]
     
+        self.n_jobs = cpu_count() if multiprocessing else 1
+
     def set_scoped_indices(self, indices):
         indices = list(indices)
         assert all_el_unique(indices), f'all scoped batch indices must be unique, received: {indices}'
@@ -177,14 +187,20 @@ class KNNMemory():
         num_memories = memories.shape[1]
 
         knn_insert_ids = np.arange(num_memories)
+
         keys = np.ascontiguousarray(memories[..., 0, :])
+        knns = [self.knns[i] for i in self.scoped_indices]
+        db_offsets = [self.db_offsets[i] for i in self.scoped_indices]
 
-        for key, index in zip(keys, self.scoped_indices):
-            db_offset = self.db_offsets[index]
-            knn = self.knns[index]
+        # use joblib to insert new key / value memories into faiss index
 
+        def process(knn, key, db_offset):
             knn.add(key, ids = knn_insert_ids + db_offset)
 
+        Parallel(n_jobs = self.n_jobs)(delayed(process)(*args) for args in zip(knns, keys, db_offsets))
+
+        # add the new memories to the memmap "database"
+
         add_indices = (rearrange(np.arange(num_memories), 'j -> 1 j') + rearrange(self.db_offsets[list(self.scoped_indices)], 'i -> i 1')) % self.max_memories
         self.db[rearrange(np.array(self.scoped_indices), 'i -> i 1'), add_indices] = memories
         self.db.flush()
